import json

from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.forms.jsonb import (
    InvalidJSONInput,
    JSONField as JSONFormField,
)

__all__ = ('UTF8JSONFormField', 'UTF8JSONField')


class UTF8JSONFormField(JSONFormField):

    def prepare_value(self, value):
        if isinstance(value, InvalidJSONInput):
            return value
        return json.dumps(value, ensure_ascii=False)


class UTF8JSONField(JSONField):
    """JSONField for postgres databases.

    Displays UTF-8 characters directly in the admin, i.e. äöü instead of
    unicode escape sequences.
    """

    def formfield(self, **kwargs):
        return super().formfield(**{
            **{'form_class': UTF8JSONFormField},
            **kwargs,
        })



import re
from decimal import Decimal
from urllib.parse import urlparse, parse_qs

UTM_PARAMS = ['utm_source', 'utm_medium', 'utm_campaign', 'utm_term', 'utm_content', 'utm_referrer']

__all__ = ('parse_utm', 'get_operations', 'calc', 'calculate', 'remove_tags', 'replacer')


def parse_utm(url: str) -> dict:
    """
    :param url: str
    :return: dict
    """
    out = {}
    try:
        url_params = parse_qs(urlparse(url.replace('#', '')).query)
        for utm in UTM_PARAMS:
            if utm in url_params:
                out[utm] = url_params[utm][0]
        return out
    except (KeyError, ValueError, IndexError, TypeError):
        return {}


def get_operations(data: dict) -> dict:
    """
    create valid calculation dict
    :param data: dict
    :return: dict
    """
    if data:
        if "cogs" in data or "crm_net_cost" in data:
            data["first_cost"] = data.pop('cogs') if "cogs" in data else data.pop('crm_net_cost')
        if "revenue" in data or "crm_deal_cost" in data:
            data['transaction_amount'] = data.pop('revenue') if "revenue" in data else data.pop('crm_deal_cost')
        return {field: str(data[field]).replace('{', '').replace('}', '') for field in data if data[field]}
    return {}


def calc(s: str) -> str:
    """
    :param s: str (value like ('(1 * 2) / 3'))
    :return: str (result of math operation)
    """
    val = s.group()
    if not val.strip():
        return val
    return "%s" % eval(val.strip(), {'__builtins__': None})


def calculate(s: str) -> str:
    """
    :param s: str
    :return: str
    """
    return re.sub(r"([0-9\ \.\+\*\-\/(\)]+)", calc, s)


def replacer(string: str, data: dict) -> str:
    """
    replace for calculated data
    :param string: str
    :param data: str
    :return: str
    """
    for name, value in data.items():
        if name in string:
            string = string.replace(name, Decimal(value).__str__())
    return string


def remove_tags(text: str) -> str:
    """
    remove html tags from text
    :param text: str
    :return: str
    """
    TAG_RE = re.compile(r'<[^>]+>')
    return TAG_RE.sub('', text)



from django.db.models import Manager
from dynamicfield_serializer import DynamicFieldSerializer
from rest_framework.serializers import ListSerializer, LIST_SERIALIZER_KWARGS

from .utils import replacer, calculate, get_operations

__all__ = ('CalculationDynamicFieldSerializer', 'GeneratorSerializer', 'GeneratorListSerializer')


class GeneratorListSerializer(ListSerializer):

    def to_representation(self, data, to_list=True):
        iterable = data.iterator() if isinstance(data, Manager) else data
        if to_list:
            return [self.child.to_representation(item) for item in iterable]
        return (self.child.to_representation(item) for item in iterable)

    @property
    def data(self):
        return self.to_representation(self.instance, to_list=True)

    @property
    def generator(self):
        return self.to_representation(self.instance, to_list=False)


class GeneratorSerializer(DynamicFieldSerializer):
    """if many=True, have property generator - to return generator serializable values"""
    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        return GeneratorListSerializer(*args, **list_kwargs)


class CalculationDynamicFieldSerializer(GeneratorSerializer):
    def to_representation(self, instance):
        ret = super().to_representation(instance)
        calc = self.context.get("calculation")
        if calc:
            calc = get_operations(calc)
            for name, value in ret.items():
                if name in calc:
                    data = replacer(calc[name], ret)
                    ret[fname] = calculate(data)

        return ret



import requests
from urllib.parse import unquote

from rest_framework.exceptions import PermissionDenied
from django.conf import settings

__all__ = ('CredentialMixin', )


class CredentialMixin(object):
    if settings.DASHBOARD_URL.endswith("/"):
        url = settings.DASHBOARD_URL
    else:
        url = settings.DASHBOARD_URL + '/'

    def get_user_info(self, request, get_token=False):
        if get_token:
            token = unquote(request.GET["token"])
        else:
            if 'HTTP_AUTHORIZATION' not in request.META or "platform_id" not in request.GET:
                raise PermissionDenied({"status": "error", "message": "you don't have permission to access"})
            token = request.META["HTTP_AUTHORIZATION"]
        try:
            response = requests.get(f"{self.url}api/check-platform/?platform_id={request.GET['platform_id']}",
                                    headers={"executetoken": token})
        except requests.ConnectionError:
            raise PermissionDenied({"status": "error", "message": "Authorization failed."})

        if response.status_code != 200:
            raise PermissionDenied({"status": "error", "message": "Authorization failed."})

        return response.json()["auth_info"]

    # method for dashboard app
    def get_check_dash_auth(self, request):
        if 'HTTP_EXECUTETOKEN' not in request.META:
            raise PermissionDenied({"status": "error", "message": "you don't have permission to access"})
        auth_info = requests.post(f"{self.url}api/check-auth/",
                                  headers={"executetoken": request.META['HTTP_EXECUTETOKEN']}).json()
        if "is_admin" not in auth_info:
            raise PermissionDenied({"status": "error", "message": "Invalid dash user."})

        return auth_info["is_admin"]

    def get_check_admin(self, request, *args, **kwargs):
        user_info = self.get_user_info(request, *args, **kwargs)
        user_id = user_info["id"]
        if user_id in (19, 22, 23, 55):
            return user_id
        return False

    def get_credential(self, request, model, check_user=True):
        user_info = self.get_user_info(request)
        user_id = user_info["id"]
        cred_id = request.GET.get("credential_id")

        params = {
            'id': cred_id,
        }
        if check_user:
            params['user_id'] = user_id

        credential = model.objects.filter(**params).first()

        if not credential:
            raise PermissionDenied({"status": "error", "message": "Credential does not exist."})

        return credential



import requests
import datetime

from django.http import HttpResponseRedirect
from rest_framework.views import APIView
from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response
from rest_framework.exceptions import PermissionDenied, NotAcceptable

from .mixins import CredentialMixin
from .pagination import LimitOffsetPaginationListAPIView

__all__ = ('HomeAPIView', 'BaseCredentialModelViewSet', 'BaseReportListAPIView', 'BaseGetCredentialAPIView')


class BaseCredentialModelViewSet(CredentialMixin, ModelViewSet):
    queryset = None
    serializer_class = None

    def get_queryset(self):
        if not self.queryset:
            raise NotImplementedError("model must be declare")
        return self.queryset

    def create(self, request, format=None):
        raise NotImplementedError()

    def retrieve(self, request, *args, **kwargs):
        """
        ---
        parameters:
            - Authorization token in headers
        """

        _ = self.get_check_dash_auth(request)
        return super().retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        """
        ---
        parameters:
            - Authorization token in headers
        """
        _ = self.get_check_dash_auth(request)

        return super().list(request, *args, **kwargs)


# base View model
class BaseReportListAPIView(LimitOffsetPaginationListAPIView):
    serializer_class = None
    model = None
    select_related_fields = ()  # must be tuple or list
    with_platforms_sample = False

    def get_model(self):
        if not self.model:
            raise NotImplementedError("model must be declare")
        return self.model

    def get_queryset(self):
        if self.select_related_fields and isinstance(self.select_related_fields, (tuple, list)):
            qs = self.get_model().objects.select_related(
                *self.select_related_fields
            ).filter(integration_id=self.integration_id,
                     date__range=[
                         self.date_from, self.date_to
                     ])
        else:
            qs = self.get_model().objects.filter(integration_id=self.integration_id,
                                                 date__range=[
                                                     self.date_from, self.date_to
                                                 ])
        if self.with_platforms_sample:
            return qs.filter(platform_id=self.platform_id)
        return qs


class BaseGetCredentialAPIView(CredentialMixin, APIView):

    def get_redirect_uri(self):
        """Create in this method redirect uri
            Example: https://oauth.yandex.ru/authorize?response_type=code&client_id={settings.YANDEX_APP_CLIENT_ID}&state={self.get_state()}&redirect_uri={redirect_url}
        """
        raise NotImplementedError()

    def check_auth_params(self):
        """This method for check auth params"""
        if "callback_url" not in self.request.GET \
                or "token" not in self.request.GET \
                or "platform_id" not in self.request.GET:
            raise PermissionDenied({"status": "error", "message": "You dont have permission to access."})
        return self.request

    def get_state(self):
        """
            Create state in this method
            Excample:
                user_info = self.get_user_info(request, get_token=True)
                user_id = user_info["id"]
                main_user = user_info["username"]
                # user_id = 16
                # main_user = "main_user"

                state = {
                    "callback_url": request.GET['callback_url'],
                    "user_id": user_id,
                    "main_user": main_user,
                    "platform_id": request.GET['platform_id']
                }
                return state
        """
        raise NotImplemented()

    def get(self, request, format=None):
        """
        ---
        parameters:
            - Authorization token in headers
            - 'callback_url' in GET params
        """

        return HttpResponseRedirect(self.get_redirect_uri())


class HomeAPIView(CredentialMixin, APIView):
    def get(self, request, format=None):
        return Response({"status": "success", "message": "smart integration api"})



__all__ = ('validate_data',)


class __Validate(object):
    def __init__(self, filter_str: str, validate_obj: dict):
        self.parsed_filters = self.__parse_filters(filter_str)
        self.validate_obj = validate_obj

    def __parse_filters(self, filter_str: str):
        """
        :param filter_:
        :return: list of dicts/lists
        """
        parsed_filters = []
        if filter_str:
            filter_values = filter_str.split(",")
            for filters in filter_values:
                if filters:
                    if ";" in filters:  # check 'AND' statement
                        and_filters = []
                        for f in filters.split(";"):
                            and_filters.append(self.__create_filter_row(f.strip()))  # strip string for del spaces
                        parsed_filters.append(and_filters)  # return list of AND filters
                        continue
                    parsed_filters.append(self.__create_filter_row(filters.strip()))  # strip string for del spaces
        return parsed_filters

    def generate_mongo_query(self, parsed: dict) -> dict:
        query = {}
        operator = parsed['set']
        field = parsed['field']
        exclude = field.startswith('!')
        field = field.replace('!', '')
        if operator == 'in_':
            query[f'{field}__regex' if not exclude else f'{field}__regex_ne'] = parsed['value']
        elif operator == 'not_in':
            query[f'{field}__regex_ne' if not exclude else f'{field}__regex'] = parsed['value']
        elif operator == 'start':
            query[f'{field}__startswith' if not exclude else f'{field}__not_startswith'] = parsed['value']
        elif operator == 'end':
            query[f'{field}__endswith' if not exclude else f'{field}__not_endswith'] = parsed['value']
        elif operator == 'eq':
            query[f'{field}' if not exclude else f'{field}__ne'] = parsed['value']
        elif operator == 'not_eq':
            query[f'{field}__ne' if not exclude else f'{field}'] = parsed['value']
        return query

    def generate_query(self, mongo: bool = True) -> list:
        queries = []
        for parsed in self.parsed_filters:
            if isinstance(parsed, list):
                query_dict = {} if mongo else {'filter': {}, 'exclude': {}}
                for p in parsed:
                    if mongo:
                        query_dict.update()
                    else:
                        f, ex = self.generate_mongo_query(p)
                        query_dict['filter'].update(f)
                        query_dict['exclude'].update(ex)
                queries.append(query_dict)
            else:
                queries.append(
                    self.generate_mongo_query(parsed) if mongo else self.generate_django_orm_query(parsed)
                )
        return queries

    def generate_django_orm_query(self, parsed: dict) -> dict:
        query = {'filter': {}, 'exclude': {}}
        operator = parsed['set']
        field = parsed['field']
        exclude = field.startswith('!')
        field = field.replace('!', '')
        if operator == 'in_':
            query['filter' if not exclude else 'exclude'][f'{field}__iexact'] = parsed['value']
        elif operator == 'not_in':
            query['exclude' if not exclude else 'filter'][field] = parsed['value']
        elif operator == 'start':
            query['filter' if not exclude else 'exclude'][f'{field}__startswith'] = parsed['value']
        elif operator == 'end':
            query['filter' if not exclude else 'exclude'][f'{field}__endswith'] = parsed['value']
        elif operator == 'eq':
            query['filter' if not exclude else 'exclude'][field] = parsed['value']
        elif operator == 'not_eq':
            query['exclude' if not exclude else 'filter'][field] = parsed['value']
        return query

    # create one query for parse_filters
    def __create_filter_row(self, f: str) -> dict:
        """
        :param f: string
        :return: dict
        """
        operators = {
            "=@": "in_",
            "!@": "not_in",
            "=$": "end",
            "=^": "start",
            "==": "eq",
            "!=": "not_eq"
        }
        for operator in operators:
            if operator in f:
                set = operators[operator]
                field, value = f.split(operator)
                return {"field": field, "value": value, "set": set}
        else:
            raise ValueError("invalid operator")

    def _validate_row(self, row: dict):
        exclude = False
        exist = False
        if row['field'].startswith('!'):
            exclude = True
            row['field'] = row['field'][1:]
        method = getattr(self, row['set'])
        check_value = self.validate_obj.get(row['field'])
        if check_value:
            exist = method(row['value'], check_value)
        return self.create_filter_result(exist, exclude)

    def validate_data(self):
        validated_result = []
        for filter_row in self.parsed_filters:
            if isinstance(filter_row, list):
                result = [self._validate_row(row) for row in filter_row]
                validated_result.append(all(result))
            else:
                validated_result.append(self._validate_row(filter_row))
        return any(validated_result)

    def in_(self, value, check_value):
        return value in check_value

    def not_in(self, value, check_value):
        return value not in check_value

    def start(self, value, check_value):
        return check_value.startswith(value)

    def end(self, value, check_value):
        return check_value.endswith(value)

    def eq(self, value, check_value):
        return check_value == value

    def not_eq(self, value, check_value):
        return check_value != value

    def create_filter_result(self, exist: bool, exclude: bool):
        """
        :param exist bool
        :param exclude bool
        :return: bool
        """
        return exist ^ exclude


def validate_data(filter_str: str, validate_obj: dict) -> bool:
    return __Validate(filter_str, validate_obj).validate_data()


def parse_for_django_query(filter_str: str) -> list:
    """
    :param filter_str: filter string
    :return: dict like {'filter': {'url__iexact': 'test.io'}, 'exclude': {'url__iexact': 'us.test.io'}}
    """
    return __Validate(filter_str, {}).generate_query(mongo=False)


def parse_for_mongo_query(filter_str: str) -> list:
    """
    :param filter_str: filter string
    :return: dict (dict for mongodantic odm)
    """
    return __Validate(filter_str, {}).generate_query(mongo=True)



from datetime import datetime, timedelta
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.generics import ListAPIView
from rest_framework.response import Response
from rest_framework.exceptions import NotAcceptable, PermissionDenied

from .mixins import CredentialMixin

__all__ = ('LimitOffsetPaginationListAPIView',)


class StandardResultsPagination(LimitOffsetPagination):
    default_limit = 1000
    max_limit = 1000


class LimitOffsetPaginationListAPIView(CredentialMixin, ListAPIView):
    """
    limit=1000
    offset=1000
    for calculation:
        rewrite get_calculation method -> dict
    for dynamic fields:
        (Need DynamicFieldSerializer)
        rewrite get_fields  method -> list
    you always have:
    self.integration_id
    self.date_form
    self.date_to
    """
    limit = None
    offset = None
    pagination_class = StandardResultsPagination

    def _ensure_limit_offset_params(self):
        if self.limit and self.offset:
            self.pagination_class.limit = limit
            self.pagination_class.offset = offset

    def _generate_date_params(self, delta: int) -> tuple:
        date_to = datetime.today().strftime('%Y-%m-%d')
        date_from = (datetime.today() - timedelta(days=delta)).strftime('%Y-%m-%d')
        return date_from, date_to

    def _get_params(self):
        admin = self.get_check_admin(self.request)
        if not admin:
            raise PermissionDenied({"status": "error", "message": "this method for admin"})

        self._ensure_limit_offset_params()
        self.integration_id = self.request.GET.get("integration_id")
        last_week = self.request.GET.get("last_week", "")
        last_3_days = self.request.GET.get("last_3_days", "")
        date_from = self.request.GET.get("date_from", "")
        date_to = self.request.GET.get("date_to", "")

        if not self.integration_id:
            raise NotAcceptable({"status": "error", "message": "miss integration_id"})

        if last_week == "true":
            date_from, date_to = self._generate_date_params(7)
        elif last_3_days == "true":
            date_from, date_to = self._generate_date_params(3)

        if not date_from and not date_to:
            raise NotAcceptable({"status": "error", "message": "miss date params"})
        self.date_from = date_from
        self.date_to = date_to
        self.platform_id = self.request.GET['platform_id']

    def get_calculation(self) -> dict:
        return {}

    def get_fields(self) -> list:
        return []

    def list(self, request, format=None):
        self._get_params()
        queryset = self.get_queryset()
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True, context={"calculation": self.get_calculation(),
                                                                       "response_fields": self.get_fields()})
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(
            queryset,
            many=True,
            context={"calculation": self.get_calculation(), "response_fields": self.get_fields()}
        )

        return Response(serializer.data)



from setuptools import setup, find_packages
from integration_utils import __version__

setup(
    name='smart-integration-utils',
    version=__version__,
    packages=find_packages(),
    setup_requires=['Django>=2.2.7',
                    'djangorestframework>=3.8.2',
                    'requests>=2.18.2',
                    ],
    install_requires=[
        'psycopg2-binary>=2.7.4',
        'pytz>=2018.4',
        'six>=1.11.0',
        'eventmonitoring-client',
        'config_field',
        'drf-dynamicfieldserializer',
        'pycrypto==2.6.1',
    ],
    python_requires='>=3.7'
)



import json
import os
import csv
import boto3
import requests
import hashlib
from io import StringIO, BytesIO
from boto3.s3.transfer import TransferConfig
from Crypto.Cipher import AES
from base64 import b64encode, b64decode
from typing import Union, Optional, List, Iterator, Generator

from django.conf import settings

BUCKET_NAME = settings.BUCKET_NAME
if settings.DASHBOARD_URL.endswith("/"):
    DASHBOARD_URL = settings.DASHBOARD_URL
else:
    DASHBOARD_URL = settings.DASHBOARD_URL + "/"

__all__ = ('chunk_of_generators', 'get_smart_token', 'delete_file', 'write_file', 'upload_to_s3', 'check_file_exist',
           'get_file_link', 'chunk_by_items', 'delete_s3_file', 'chunk_by_parts', 'decode', 'encode',
           'get_full_integrations', 'write_ram_file', 'upload_to_s3_from_ram', 'delete_invalid_integrations')


def chunk_of_generators(iterable: Union[Generator, Iterator], max_value: int = 10):
    def inner(iterator: Union[Generator, Iterator], max_value: int, switcher: list):
        switcher[0] = False
        for i in range(max_value):
            try:
                yield next(iterator)
            except StopIteration:
                return
        switcher[0] = True
        return

    iterator = iter(iterable)
    switcher = [True]
    while switcher[0] == True:
        try:
            yield inner(iterator, max_value, switcher)
        except StopIteration:
            return None


def get_smart_token(platform: str) -> str:
    """
    :param platform: str (prod or dev)
    :return: str
    """
    url = f'{DASHBOARD_URL}api/smart_tokens/?platform_id={platform}'
    token = settings.EXECUTE_TOKEN
    try:
        r = requests.get(url, headers={"executetoken": token}).json()
    except (requests.ConnectionError, json.JSONDecodeError):
        return 'error'
    return r['token']


def delete_file(filename: str):
    os.remove(filename)


def write_file(filename: str, data: Union[list, dict, Generator], fieldnames: Optional[List[str]] = None,
               delimiter: str = '\t', extension: str = '.tsv') -> str:
    """
    :param filename: str (path to local file)
    :param data: any (serialized data)
    :param delimiter: str
    :param extension: str
    :param fieldnames: list or None (list of fieldsnames)
    :return: bool
    """
    if extension not in filename:
        filename = f"{filename}{extension}"
    with open(filename, 'w') as f:
        if not fieldnames:
            fieldnames = list(data[0].keys())
        writer = csv.DictWriter(f, delimiter=delimiter, fieldnames=fieldnames)
        writer.writeheader()
        for row in data:
            writer.writerow(row)
    return filename


def upload_to_s3(filename: str, folder: str = "stats/") -> dict:
    """
    :param filename: str (path to local file)
    :param folder: str (folder from s3)
    :return: dict
    """
    print(f"=== IN BOTO3 {filename}")
    s3 = boto3.resource('s3')
    GB = 1024 ** 3
    config = TransferConfig(multipart_threshold=int(GB / 3))
    if filename.startswith('/tmp/'):
        s3_filename = filename[5:]
    else:
        s3_filename = filename
    s3.meta.client.upload_file(filename, BUCKET_NAME, f"{folder}{s3_filename}", ExtraArgs={'ACL': 'public-read'},
                               Config=config)

    return {
        "status": "success",
        "s3_filename": s3_filename,
        "s3_folder": folder
    }


def check_file_exist(filename: str, folder: str = "stats/") -> bool:
    """
    :param filename: str
    :param folder: str
    :return: bool
    """
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(BUCKET_NAME)
    if filename.startswith('/tmp/'):
        filename = filename[5:]
    key = folder + filename
    objs = list(bucket.objects.filter(Prefix=key))
    if len(objs) > 0 and objs[0].key == key:
        return True
    else:
        return False


def get_file_link(filename: str, folder: str = 'stats/', bucket: str = BUCKET_NAME, suffix: str = '.tsv') -> str:
    """
    :param filename: str
    :param folder: str
    :param bucket:  str
    :param suffix: str
    :return: str
    """
    client = boto3.client('s3')
    if filename.startswith('/tmp/'):
        filename = filename[5:]
    if suffix in filename:
        filename = f"{folder}{filename}" if folder not in filename else filename
    else:
        filename = f"{folder}{filename}{suffix}" if folder not in filename else f"{filename}{suffix}"
    # content = client.head_object(Bucket=BUCKET_NAME, Key=filename)
    bucket_location = client.get_bucket_location(Bucket=bucket)
    filepath = f"https://s3-{bucket_location['LocationConstraint']}.amazonaws.com/{BUCKET_NAME}/{filename}"

    return filepath


def chunk_by_items(items: list, step: int):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(items), step):
        yield items[i: i + step]


def get_full_integrations(platform: str, service_code: str) -> list:
    """
    :param platform: str (prod or dev)
    :param service_code: str (code from integration config)
    :return: list (list of integrations)
    """
    token = get_smart_token(platform)
    url = f'{DASHBOARD_URL}api/get_config_url/'
    dash_response = requests.get(f'{url}?platform_id={platform}').json()
    headers = {"Authorization": token}

    config_url = f'{dash_response["config_url"]}?type__code={service_code}&all=1'
    # print(config_url)

    configs = requests.get(config_url, headers=headers).json()
    results = configs["results"]
    while configs["next"]:
        configs = requests.get(configs["next"], headers=headers).json()
        results.extend(configs["results"])
    return results


def check_entities_equal(db_entity: any, new_entity: any) -> bool:
    """
    :param db_entity: any
    :param new_entity: any
    :return: bool
    """
    db_entity = db_entity.__dict__
    new_entity = new_entity.__dict__
    exclude_fields = ('_state', 'date_time', 'sc_date_time', 'funnel_stage_duration', 'id')
    return all((field, new_entity[field]) in db_entity.items() for field in db_entity if field not in exclude_fields)


def chunk_by_parts(seq: list, parts: int) -> list:
    """
    :param seq: list
    :param parts: int
    :return: list
    """
    if len(seq) < parts:
        return [seq]
    avg = len(seq) / float(parts)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last):int(last + avg)])
        last += avg
    return out


def delete_s3_file(filename: str, folder: str, suffix: str) -> str:
    """
    :param filename: str (like /tmp/yandex_stats)
    :param folder: str (s3 folder name like stats/)
    :param suffix: str (end of file like .csv)
    :return: str
    """
    if filename.startswith('/tmp/'):
        filename = filename[5:]
    client = boto3.client('s3')
    if suffix not in filename:
        filename = f"{filename}{suffix}"
    if folder not in filename:
        filename = f"{folder}{filename}"
    client.delete_object(Bucket=BUCKET_NAME, Key=filename)
    return f"{filename} - deleted!"


class __Crypt(object):
    def __init__(self, salt='SlTKeYOpHygTYkP3'):
        self.salt = salt
        self.enc_dec_method = 'utf-8'

    def encrypt(self, str_to_enc, str_key=settings.SECRET_KEY):
        try:
            aes_obj = AES.new(hashlib.sha256(str_key.encode()).digest(), AES.MODE_CFB, self.salt)
            hx_enc = aes_obj.encrypt(str_to_enc)
            mret = b64encode(hx_enc).decode(self.enc_dec_method)
            mret = mret.replace('/', '-_-')
            return mret
        except ValueError as value_error:
            if value_error.args[0] == 'IV must be 16 bytes long':
                raise ValueError('Encryption Error: SALT must be 16 characters long')
            elif value_error.args[0] == 'AES key must be either 16, 24, or 32 bytes long':
                raise ValueError('Encryption Error: Encryption key must be either 16, 24, or 32 characters long')
            else:
                raise ValueError(value_error)

    def decrypt(self, enc_str, str_key=settings.SECRET_KEY):
        try:
            enc_str = enc_str.replace('-_-', '/')
            aes_obj = AES.new(hashlib.sha256(str_key.encode()).digest(), AES.MODE_CFB, self.salt)
            str_tmp = b64decode(enc_str.encode(self.enc_dec_method))
            str_dec = aes_obj.decrypt(str_tmp)
            mret = str_dec.decode(self.enc_dec_method)
            return mret
        except ValueError as value_error:
            if value_error.args[0] == 'IV must be 16 bytes long':
                raise ValueError('Decryption Error: SALT must be 16 characters long')
            elif value_error.args[0] == 'AES key must be either 16, 24, or 32 bytes long':
                raise ValueError('Decryption Error: Encryption key must be either 16, 24, or 32 characters long')
            else:
                raise ValueError(value_error)


def encode(string: str, secret: Optional[str] = None):
    if not secret:
        secret = settings.SECRET_KEY
    c = __Crypt()
    return c.encrypt(string, secret)


def decode(enc: str, secret: Optional[str] = None):
    if not secret:
        secret = settings.SECRET_KEY
    c = __Crypt()
    return c.decrypt(enc, secret)


def write_ram_file(f: StringIO, data: Generator, fieldnames: list, delimiter='\t') -> StringIO:
    writer = csv.DictWriter(f, delimiter=delimiter, fieldnames=fieldnames)
    writer.writeheader()
    for row in data:
        writer.writerow(row)
    return f


def upload_to_s3_from_ram(content: Union[StringIO, BytesIO], filename: str, folder: str = 'yandex_files/',
                          suffix: str = '.tsv'):
    if filename.startswith('/tmp/'):
        s3_filename = filename[5:]
    else:
        s3_filename = filename
    if suffix not in s3_filename:
        s3_filename = f"{s3_filename}{suffix}"
    s3 = boto3.resource('s3')
    body = content.getvalue().encode('utf-8') if isinstance(content, StringIO) else content.getvalue()
    s3.meta.client.put_object(Bucket=BUCKET_NAME, Key=f'{folder}{s3_filename}',
                              Body=body, ACL='public-read')
    return {
        "status": "success",
        "s3_filename": s3_filename,
        "s3_folder": folder
    }


def delete_invalid_integrations(integrations: list, models: Union[tuple, list],
                                filter_query: Optional[dict] = None) -> list:
    integration_ids = [i['integration_id'] for i in integrations]
    return [
        model.objects.exclude(integration_id__in=integration_ids).delete() if not filter_query else
        model.objects.filter(**filter_query
                             ).exclude(integration_id__in=integration_ids).delete()
        for model in models
    ]
